@@ -1,56 +1,58 @@
 #include "common.h"
+#include "pathutils.h"
 
-static int progress (u_int64_t const sent, u_int64_t const total, void const * const data)
-{
-  int percent = (sent*100)/total;
-#ifdef __WIN32__
-  printf("Progress: %I64u of %I64u (%d%%)\r", sent, total, percent);
-#else
-  printf("Progress: %llu of %llu (%d%%)\r", sent, total, percent);
-#endif
-  fflush(stdout);
-  return 0;
-}
+void getfile_function(char *,char *);
+void getfile_command(int, char **);
+void getfile_usage(void);
 
-static void usage (void)
+extern LIBMTP_folder_t *folders;
+extern LIBMTP_file_t *files;
+extern LIBMTP_mtpdevice_t *device;
+
+void getfile_usage (void)
 {
   fprintf(stderr, "getfile <fileid/trackid> <filename>\n");
 }
 
-int main (int argc, char **argv)
+void
+getfile_function(char * from_path,char * to_path)
+{
+  int id = parse_path (from_path,files,folders);
+  if (id > 0) {
+    printf("Getting %s to %s\n",from_path,to_path);
+    if (LIBMTP_Get_File_To_File(device, id, to_path, progress, NULL) != 0 ) {
+      printf("\nError getting file from MTP device.\n");
+    }
+  }
+}
+
+
+void getfile_command(int argc, char **argv)
 {
-  LIBMTP_mtpdevice_t *device;
   u_int32_t id;
   char *endptr;
   char *file;
 
   // We need file ID and filename
   if ( argc != 3 ) {
-    usage();
-    return 1;
+    getfile_usage();
+    return;
   }
 
   // Sanity check song ID
   id = strtoul(argv[1], &endptr, 10);
   if ( *endptr != 0 ) {
     fprintf(stderr, "illegal value %s\n", argv[1]);
-    return 1;
+    return;
   } else if ( ! id ) {
     fprintf(stderr, "bad file/track id %u\n", id);
-    return 1;
+    return;
  }
 
   // Filename, e.g. "foo.mp3"
   file = argv[2];
   printf("Getting file/track %d to local file %s\n", id, file);
 
-  LIBMTP_Init();
-  device = LIBMTP_Get_First_Device();
-  if (device == NULL) {
-    printf("No devices. Connect/replug device and try again.\n");
-    exit (0);
-  }
-  
   // This function will also work just as well for tracks.
   if (LIBMTP_Get_File_To_File(device, id, file, progress, NULL) != 0 ) {
     printf("\nError getting file from MTP device.\n");
@@ -58,8 +60,6 @@ int main (int argc, char **argv)
   // Terminate progress bar.
   printf("\n");
   
-  LIBMTP_Release_Device(device);
-  printf("OK.\n");
-  exit (0);
+  return;
 }
 
