@@ -151,40 +151,36 @@ ptp_pack_string(PTPParams *params, char *string, unsigned char* data, uint16_t o
 	ucs2str[0] = 0x0000U;
 	memset(ucs2strp, 0, PTP_MAXSTRLEN*2+2);
 	nconv = iconv (params->cd_locale_to_ucs2, &stringp, &convlen, &ucs2strp, &convmax);
-	if (nconv == (size_t) -1) {
+	if (nconv == (size_t) -1)
 		ucs2str[0] = 0x0000U;
-	}
 	packedlen = ucs2strlen(ucs2str);
 	if (packedlen > PTP_MAXSTRLEN-1) {
 		*len=0;
 		return;
 	}
 	
-	/* +1 for the length byte, no zero 0x0000 terminator */
+	/* number of characters including terminating 0 (PTP standard confirmed) */
 	htod8a(&data[offset],packedlen+1);
 	for (i=0;i<packedlen && i< PTP_MAXSTRLEN; i++) {
 		htod16a(&data[offset+i*2+1],ucs2str[i]);
 	}
+	htod16a(&data[offset+i*2+1],0x0000);
 
 	/* The returned length is in number of characters */
-	*len = (uint8_t) packedlen;
+	*len = (uint8_t) packedlen+1;
 }
 
 static inline unsigned char *
 ptp_get_packed_stringcopy(PTPParams *params, char *string, uint32_t *packed_size)
 {
-uint8_t packed[PTP_MAXSTRLEN+3], len;
+	uint8_t packed[PTP_MAXSTRLEN*2+3], len;
 	size_t plen;
 	unsigned char *retcopy = NULL;
 
 	ptp_pack_string(params, string, (unsigned char*) packed, 0, &len);
 	/* returned length is in characters, then one byte for string length */
 	plen = len*2 + 1;
-	/* Assure proper termination, two zero bytes */
-	packed[plen] = 0x00;
-	packed[plen+1] = 0x00;
-	/* Include terminator */
-	plen += 2;
+	
 	retcopy = malloc(plen);
 	if (!retcopy) {
 		*packed_size = 0;
@@ -250,7 +246,7 @@ ptp_unpack_DI (PTPParams *params, unsigned char* data, PTPDeviceInfo *di, unsign
 	uint8_t len;
 	unsigned int totallen;
 	
-	di->StaqndardVersion = dtoh16a(&data[PTP_di_StandardVersion]);
+	di->StandardVersion = dtoh16a(&data[PTP_di_StandardVersion]);
 	di->VendorExtensionID =
 		dtoh32a(&data[PTP_di_VendorExtensionID]);
 	di->VendorExtensionVersion =
@@ -368,19 +364,23 @@ ptp_unpack_SI (PTPParams *params, unsigned char* data, PTPStorageInfo *si, unsig
 #define PTP_oi_filenamelen		52
 #define PTP_oi_Filename			53
 
+/* the max length assuming zero length dates. We have need 3 */
+/* bytes for these. */
+#define PTP_oi_MaxLen PTP_oi_Filename+(PTP_MAXSTRLEN+1)*2+3
+
 static inline uint32_t
 ptp_pack_OI (PTPParams *params, PTPObjectInfo *oi, unsigned char** oidataptr)
 {
 	unsigned char* oidata;
 	uint8_t filenamelen;
 	uint8_t capturedatelen=0;
-	/* let's allocate some memory first; XXX i'm sure it's wrong */
-	oidata=malloc(PTP_oi_Filename+(strlen(oi->Filename)+1)*2+4);
+	/* let's allocate some memory first; correct assuming zero length dates */
+	oidata=malloc(PTP_oi_MaxLen);
 	/* the caller should free it after use! */
 #if 0
 	char *capture_date="20020101T010101"; /* XXX Fake date */
 #endif
-	memset (oidata, 0, (PTP_oi_Filename+(strlen(oi->Filename)+1)*2+4));
+	memset (oidata, 0, PTP_oi_MaxLen);
 	htod32a(&oidata[PTP_oi_StorageID],oi->StorageID);
 	htod16a(&oidata[PTP_oi_ObjectFormat],oi->ObjectFormat);
 	htod16a(&oidata[PTP_oi_ProtectionStatus],oi->ProtectionStatus);
@@ -427,7 +427,7 @@ ptp_pack_OI (PTPParams *params, PTPObjectInfo *oi, unsigned char** oidataptr)
 	/* XXX this function should return dataset length */
 	
 	*oidataptr=oidata;
-	return (PTP_oi_Filename+(filenamelen+1)*2+(capturedatelen+1)*4);
+	return (PTP_oi_Filename+filenamelen*2+(capturedatelen+1)*3);
 }
 
 static inline void
@@ -562,6 +562,28 @@ ptp_unpack_DPV (
 	case PTP_DTC_UINT32:
 		CTVAL(value->u32,dtoh32a);
 		break;
+
+
+
+	case PTP_DTC_UINT64:
+		*offset += 8;
+		/*fprintf(stderr,"unhandled unpack of uint64\n");*/
+		break;
+	case PTP_DTC_INT64:
+		*offset += 8;
+		/*fprintf(stderr,"unhandled unpack of int64\n");*/
+		break;
+	case PTP_DTC_UINT128:
+		*offset += 16;
+		/*fprintf(stderr,"unhandled unpack of uint128n");*/
+		break;
+	case PTP_DTC_INT128:
+		*offset += 16;
+		/*fprintf(stderr,"unhandled unpack of int128n");*/
+		break;
+
+
+
 	case PTP_DTC_AINT8:
 		RARR(value,i8,dtoh8a);
 		break;
@@ -591,6 +613,8 @@ ptp_unpack_DPV (
 			return 0;
 		break;
 	}
+	default:
+		return 0;
 	}
 	return 1;
 }
@@ -833,6 +857,106 @@ ptp_pack_DPV (PTPParams *params, PTPPropertyValue* value, unsigned char** dpvptr
 	return size;
 }
 
+#define MAX_MTP_PROPS 127
+static inline uint32_t
+ptp_pack_OPL (PTPParams *params, MTPPropList *proplist, unsigned char** opldataptr)
+{
+	unsigned char* opldata;
+	MTPPropList *propitr;
+	unsigned char *packedprops[MAX_MTP_PROPS];
+	uint32_t packedpropslens[MAX_MTP_PROPS];
+	uint32_t packedobjecthandles[MAX_MTP_PROPS];
+	uint16_t packedpropsids[MAX_MTP_PROPS];
+	uint16_t packedpropstypes[MAX_MTP_PROPS];
+	uint32_t totalsize = 0;
+	uint32_t bufp = 0;
+	uint32_t noitems = 0;
+	uint32_t i;
+
+	totalsize = sizeof(uint32_t); /* 4 bytes to store the number of elements */
+	propitr = proplist;
+	while (propitr != NULL && noitems < MAX_MTP_PROPS) {
+		/* Object Handle */
+		packedobjecthandles[noitems]=propitr->ObjectHandle;
+		totalsize += sizeof(uint32_t); /* Object ID */
+		/* Metadata type */
+		packedpropsids[noitems]=propitr->property;
+		totalsize += sizeof(uint16_t);
+		/* Data type */
+		packedpropstypes[noitems]= propitr->datatype;
+		totalsize += sizeof(uint16_t);
+		/* Add each property to be sent. */
+	        packedpropslens[noitems] = ptp_pack_DPV (params, &propitr->propval, &packedprops[noitems], propitr->datatype);
+		totalsize += packedpropslens[noitems];
+		noitems ++;
+		propitr = propitr->next;
+	}
+
+	/* Allocate memory for the packed property list */
+	opldata = malloc(totalsize);
+
+	htod32a(&opldata[bufp],noitems);
+	bufp += 4;
+
+	/* Copy into a nice packed list */
+	for (i = 0; i < noitems; i++) {
+		/* Object ID */
+		htod32a(&opldata[bufp],packedobjecthandles[i]);
+		bufp += sizeof(uint32_t);
+		htod16a(&opldata[bufp],packedpropsids[i]);
+		bufp += sizeof(uint16_t);
+		htod16a(&opldata[bufp],packedpropstypes[i]);
+		bufp += sizeof(uint16_t);
+		/* The copy the actual property */
+		memcpy(&opldata[bufp], packedprops[i], packedpropslens[i]);
+		bufp += packedpropslens[i];
+		free(packedprops[i]);
+	}
+	*opldataptr = opldata;
+	return totalsize;
+}
+
+static inline int
+ptp_unpack_OPL (PTPParams *params, unsigned char* data, MTPPropList **proplist, unsigned int len)
+{ 
+	uint32_t prop_count = dtoh32a(data);
+	MTPPropList *prop = NULL;
+	int offset = 0, i;
+
+	if (prop_count == 0) {
+		*proplist = NULL;
+		return 0;
+	}
+	data += sizeof(uint32_t);
+	*proplist = malloc(sizeof(MTPPropList));
+	prop = *proplist;
+	for (i = 0; i < prop_count; i++) {
+		prop->ObjectHandle = dtoh32a(data);
+		data += sizeof(uint32_t);
+		len -= sizeof(uint32_t);
+
+		prop->property = dtoh16a(data);
+		data += sizeof(uint16_t);
+		len -= sizeof(uint16_t);
+
+		prop->datatype = dtoh16a(data);
+		data += sizeof(uint16_t);
+		len -= sizeof(uint16_t);
+
+		offset = 0;
+		ptp_unpack_DPV(params, data, &offset, len, &prop->propval, prop->datatype);
+		data += offset;
+		len -= offset;
+
+		if (i != prop_count - 1) {
+			prop->next = malloc(sizeof(MTPPropList));
+			prop = prop->next;
+		} else
+			prop->next = NULL;
+	}
+	return prop_count;
+}
+
 /*
     PTP USB Event container unpack
     Copyright (c) 2003 Nikolai Kopanygin
@@ -966,3 +1090,140 @@ ptp_pack_EK_text(PTPParams *params, PTPEKTextParams *text, unsigned char **data)
 	}
 	return len;
 }
+
+#define ptp_canon_dir_version	0x00
+#define ptp_canon_dir_ofc	0x02
+#define ptp_canon_dir_unk1	0x04
+#define ptp_canon_dir_objectid	0x08
+#define ptp_canon_dir_parentid	0x0c
+#define ptp_canon_dir_previd	0x10	/* in same dir */
+#define ptp_canon_dir_nextid	0x14	/* in same dir */
+#define ptp_canon_dir_nextchild	0x18	/* down one dir */
+#define ptp_canon_dir_storageid	0x1c	/* only in storage entry */
+#define ptp_canon_dir_name	0x20
+#define ptp_canon_dir_flags	0x2c
+#define ptp_canon_dir_size	0x30
+#define ptp_canon_dir_unixtime	0x34
+#define ptp_canon_dir_year	0x38
+#define ptp_canon_dir_month	0x39
+#define ptp_canon_dir_mday	0x3a
+#define ptp_canon_dir_hour	0x3b
+#define ptp_canon_dir_minute	0x3c
+#define ptp_canon_dir_second	0x3d
+#define ptp_canon_dir_unk2	0x3e
+#define ptp_canon_dir_thumbsize	0x40
+#define ptp_canon_dir_width	0x44
+#define ptp_canon_dir_height	0x48
+
+static inline uint16_t
+ptp_unpack_canon_directory (
+	PTPParams		*params,
+	unsigned char		*dir,
+	uint32_t		cnt,
+	PTPObjectHandles	*handles,
+	PTPObjectInfo		**oinfos,	/* size(handles->n) */
+	uint32_t		**flags		/* size(handles->n) */
+) {
+	unsigned int	i, j, nrofobs = 0, curob = 0;
+
+#define ISOBJECT(ptr) (dtoh32a((ptr)+ptp_canon_dir_storageid) == 0xffffffff)
+	for (i=0;i<cnt;i++)
+		if (ISOBJECT(dir+i*0x4c)) nrofobs++;
+	handles->n = nrofobs;
+	handles->Handler = calloc(sizeof(handles->Handler[0]),nrofobs);
+	if (!handles->Handler) return PTP_RC_GeneralError;
+	*oinfos = calloc(sizeof((*oinfos)[0]),nrofobs);
+	if (!*oinfos) return PTP_RC_GeneralError;
+	*flags  = calloc(sizeof((*flags)[0]),nrofobs);
+	if (!*flags) return PTP_RC_GeneralError;
+
+	/* Migrate data into objects ids, handles into
+	 * the object handler array.
+	 */
+	curob = 0;
+	for (i=0;i<cnt;i++) {
+		unsigned char	*cur = dir+i*0x4c;
+		PTPObjectInfo	*oi = (*oinfos)+curob;
+
+		if (!ISOBJECT(cur))
+			continue;
+
+		handles->Handler[curob] = dtoh32a(cur + ptp_canon_dir_objectid);
+		oi->StorageID		= 0xffffffff;
+		oi->ObjectFormat	= dtoh16a(cur + ptp_canon_dir_ofc);
+		oi->ParentObject	= dtoh32a(cur + ptp_canon_dir_parentid);
+		oi->Filename		= strdup((char*)(cur + ptp_canon_dir_name));
+		oi->ObjectCompressedSize= dtoh32a(cur + ptp_canon_dir_size);
+		oi->ThumbCompressedSize	= dtoh32a(cur + ptp_canon_dir_thumbsize);
+		oi->ImagePixWidth	= dtoh32a(cur + ptp_canon_dir_width);
+		oi->ImagePixHeight	= dtoh32a(cur + ptp_canon_dir_height);
+		oi->CaptureDate		= oi->ModificationDate = dtoh32a(cur + ptp_canon_dir_unixtime);
+		(*flags)[curob]		= dtoh32a(cur + ptp_canon_dir_flags);
+		curob++;
+	}
+	/* Walk over Storage ID entries and distribute the IDs to
+	 * the parent objects. */
+	for (i=0;i<cnt;i++) {
+		unsigned char	*cur = dir+i*0x4c;
+		uint32_t	nextchild = dtoh32a(cur + ptp_canon_dir_nextchild);
+
+		if (ISOBJECT(cur))
+			continue;
+		for (j=0;j<handles->n;j++) if (nextchild == handles->Handler[j]) break;
+		if (j == handles->n) continue;
+		(*oinfos)[j].StorageID = dtoh32a(cur + ptp_canon_dir_storageid);
+	}
+	/* Walk over all objects and distribute the storage ids */
+	while (1) {
+		int changed = 0;
+		for (i=0;i<cnt;i++) {
+			unsigned char	*cur = dir+i*0x4c;
+			uint32_t	oid = dtoh32a(cur + ptp_canon_dir_objectid);
+			uint32_t	nextoid = dtoh32a(cur + ptp_canon_dir_nextid);
+			uint32_t	nextchild = dtoh32a(cur + ptp_canon_dir_nextchild);
+			uint32_t	storageid;
+
+			if (!ISOBJECT(cur))
+				continue;
+			for (j=0;j<handles->n;j++) if (oid == handles->Handler[j]) break;
+			if (j == handles->n) {
+				/*fprintf(stderr,"did not find oid in lookup pass for current oid\n");*/
+				continue;
+			}
+	 		storageid = (*oinfos)[j].StorageID;
+			if (storageid == 0xffffffff) continue;
+			if (nextoid != 0xffffffff) {
+				for (j=0;j<handles->n;j++) if (nextoid == handles->Handler[j]) break;
+				if (j == handles->n) {
+					/*fprintf(stderr,"did not find oid in lookup pass for next oid\n");*/
+					continue;
+				}
+				if ((*oinfos)[j].StorageID == 0xffffffff) {
+					(*oinfos)[j].StorageID = storageid;
+					changed++;
+				}
+			}
+			if (nextchild != 0xffffffff) {
+				for (j=0;j<handles->n;j++) if (nextchild == handles->Handler[j]) break;
+				if (j == handles->n) {
+					/*fprintf(stderr,"did not find oid in lookup pass for next child\n");*/
+					continue;
+				}
+				if ((*oinfos)[j].StorageID == 0xffffffff) {
+					(*oinfos)[j].StorageID = storageid;
+					changed++;
+				}
+			}
+		}
+		/* Check if we:
+		 * - changed no entry (nothing more to do)
+		 * - changed all of them at once (usually happens)
+		 * break if we do.
+		 */
+		if (!changed || (changed==nrofobs-1))
+			break;
+	}
+#undef ISOBJECT
+	return PTP_RC_OK;
+}
+
